##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Utils for message passing"""
import uuid
import warnings
from typing import Optional, Tuple, Union, List, Any

import rpcq.messages


def rpc_warning(warning: warnings.WarningMessage) -> rpcq.messages.RPCWarning:
    return rpcq.messages.RPCWarning(body=str(warning.message),
                                    kind=str(warning.category.__name__))

def rpc_request(method_name: str, *args, **kwargs) -> rpcq.messages.RPCRequest:
    """
    Create RPC request

    :param method_name: Method name
    :param args: Positional arguments
    :param kwargs: Keyword arguments
    :return: JSON RPC formatted dict
    """
    if args:
        kwargs['*args'] = args

    return rpcq.messages.RPCRequest(
        jsonrpc='2.0',
        id=str(uuid.uuid4()),
        method=method_name,
        params=kwargs
    )


def rpc_reply(id: Union[str, int], result: Optional[object],
              warnings: Optional[List[Warning]] = None) -> rpcq.messages.RPCReply:
    """
    Create RPC reply

    :param str|int id: Request ID
    :param result: Result
    :param warnings: List of warnings to attach to the message
    :return: JSON RPC formatted dict
    """
    warnings = warnings or []

    return rpcq.messages.RPCReply(
        jsonrpc='2.0',
        id=id,
        result=result,
        warnings=[rpc_warning(warning) for warning in warnings]
    )


def rpc_error(id: Union[str, int], error_msg: str,
              warnings: List[Any] = []) -> rpcq.messages.RPCError:
    """
    Create RPC error

    :param id: Request ID
    :param error_msg: Error message
    :param warning: List of warnings to attach to the message
    :return: JSON RPC formatted dict
    """
    return rpcq.messages.RPCError(
        jsonrpc='2.0',
        id=id,
        error=error_msg,
        warnings=[rpc_warning(warning) for warning in warnings])


def get_input(params: Union[dict, list]) -> Tuple[list, dict]:
    """
    Get positional or keyword arguments from JSON RPC params

    :param params: Parameters passed through JSON RPC
    :return: args, kwargs
    """
    # Backwards compatibility for old clients that send params as a list
    if isinstance(params, list):
        args = params
        kwargs = {}
    elif isinstance(params, dict):
        args = params.pop('*args', [])
        kwargs = params
    else:  # pragma no coverage
        raise TypeError(
            'Unknown type {} of params, must be list or dict'.format(type(params)))

    return args, kwargs


class RPCErrorError(IOError):
    """JSON RPC error that is raised by a Client when it receives an RPCError message"""

    def __init__(self, *args, **kwargs):
        if type(self) is RPCErrorError:
            warnings.warn("`RPCErrorError` is deprecated in favor of the "
                          "less-loquacious `RPCError`.", DeprecationWarning)

        super().__init__(*args, **kwargs)


class RPCError(RPCErrorError):
    """JSON RPC error that is raised by a Client when it receives an RPCError message"""


class RPCMethodError(AttributeError):
    """JSON RPC error that is raised by JSON RPC spec for nonexistent methods"""


class catch_warnings(warnings.catch_warnings):
    """This variant of warnings.catch_warnings both logs *and* re-emits warnings."""
    def __enter__(self):
        super().__enter__()

        # the super() method does most of the work.  what follows below is actually
        # also cut out of the super() method, but the relevant line there is
        #
        #     self._module._showwarnmsg_impl = log.append
        #
        # we, on the other hand, want to both append *and* call the saved parent
        # log-displayer, so we wrap both inside of new_logger and store that instead.

        if self._record:
            log = []
            def new_logger(msg):
                nonlocal log, self
                log.append(msg)
                self._showwarnmsg_impl(msg)
            self._module._showwarnmsg_impl = new_logger
            return log
        else:
            return None



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Server that accepts JSON RPC requests and returns JSON RPC replies/errors.
"""
import asyncio
import logging
from asyncio import AbstractEventLoop
from typing import Callable
from datetime import datetime

import zmq.asyncio

from rpcq._base import to_msgpack, from_msgpack
from rpcq._spec import RPCSpec
from rpcq.messages import RPCRequest

_log = logging.getLogger(__name__)


class Server:
    """
    Server that accepts JSON RPC calls through a socket.
    """
    def __init__(self, rpc_spec: RPCSpec = None, announce_timing: bool = False,
                 serialize_exceptions: bool = True):
        """
        Create a server that will be linked to a socket

        :param rpc_spec: JSON RPC spec
        :param announce_timing:
        :param serialize_exceptions: If set to True, this Server will catch all exceptions occurring
            internally to it and, when possible, communicate them to the interrogating Client.  If
            set to False, this Server will re-raise any exceptions it encounters (including, but not
            limited to, those which might occur through method calls to rpc_spec) for Server's
            local owner to handle.

            IMPORTANT NOTE: When set to False, this *almost definitely* means an unrecoverable
            crash, and the Server should then be _shutdown().
        """
        self.announce_timing = announce_timing
        self.serialize_exceptions = serialize_exceptions

        self.rpc_spec = rpc_spec if rpc_spec else RPCSpec(serialize_exceptions=serialize_exceptions)
        self._exit_handlers = []

        self._socket = None

    def rpc_handler(self, f: Callable):
        """
        Add a function to the server. It will respond to JSON RPC requests with the corresponding method name.
        This can be used as both a side-effecting function or as a decorator.

        :param f: Function to add
        :return: Function wrapper (so it can be used as a decorator)
        """
        return self.rpc_spec.add_handler(f)

    def exit_handler(self, f: Callable):
        """
        Add an exit handler - a function which will be called when the server shuts down.

        :param f: Function to add
        """
        self._exit_handlers.append(f)

    async def run_async(self, endpoint: str):
        """
        Run server main task (asynchronously).

        :param endpoint: Socket endpoint to listen to, e.g. "tcp://*:1234"
        """
        self._connect(endpoint)

        # spawn an initial listen task
        listen_task = asyncio.ensure_future(self._socket.recv_multipart())
        task_list = [listen_task]

        while True:
            dones, pendings = await asyncio.wait(task_list, return_when=asyncio.FIRST_COMPLETED)

            # grab one "done" task to handle
            task_list, done_list = list(pendings), list(dones)
            done = done_list.pop()
            task_list += done_list

            if done == listen_task:
                try:
                    # empty_frame may either be:
                    # 1. a single null frame if the client is a REQ socket
                    # 2. an empty list (ie. no frames) if the client is a DEALER socket
                    identity, *empty_frame, msg = done.result()
                    request = from_msgpack(msg)

                    # spawn a processing task
                    task_list.append(asyncio.ensure_future(
                        self._process_request(identity, empty_frame, request)))
                except Exception as e:
                    if self.serialize_exceptions:
                        _log.exception('Exception thrown in Server run loop during request '
                                       'reception: {}'.format(str(e)))
                    else:
                        raise e
                finally:
                    # spawn a new listen task
                    listen_task = asyncio.ensure_future(self._socket.recv_multipart())
                    task_list.append(listen_task)
            else:
                # if there's been an exception during processing, consider reraising it
                try:
                    done.result()
                except Exception as e:
                    if self.serialize_exceptions:
                        _log.exception('Exception thrown in Server run loop during request '
                                       'dispatch: {}'.format(str(e)))
                    else:
                        raise e

    def run(self, endpoint: str, loop: AbstractEventLoop = None):
        """
        Run server main task.

        :param endpoint: Socket endpoint to listen to, e.g. "tcp://*:1234"
        :param loop: Event loop to run server in (alternatively just use run_async method)
        """
        if not loop:
            loop = asyncio.get_event_loop()

        try:
            loop.run_until_complete(self.run_async(endpoint))
        except KeyboardInterrupt:
            self._shutdown()

    def stop(self):
        """
        DEPRECATED
        """
        pass

    def _shutdown(self):
        """
        Shut down the server.
        """
        for exit_handler in self._exit_handlers:
            exit_handler()

        if self._socket:
            self._socket.close()
            self._socket = None

    def _connect(self, endpoint: str):
        """
        Connect the server to an endpoint. Creates a ZMQ ROUTER socket for the given endpoint.

        :param endpoint: Socket endpoint, e.g. "tcp://*:1234"
        """
        if self._socket:
            raise RuntimeError('Cannot run multiple Servers on the same socket')

        context = zmq.asyncio.Context()
        self._socket = context.socket(zmq.ROUTER)
        self._socket.bind(endpoint)

        _log.info("Starting server, listening on endpoint {}".format(endpoint))

    async def _process_request(self, identity: bytes, empty_frame: list, request: RPCRequest):
        """
        Executes the method specified in a JSON RPC request and then sends the reply to the socket.

        :param identity: Client identity provided by ZeroMQ
        :param empty_frame: Either an empty list or a single null frame depending on the client type
        :param request: JSON RPC request
        """
        try:
            _log.debug("Client %s sent request: %s", identity, request)
            start_time = datetime.now()
            reply = await self.rpc_spec.run_handler(request)
            if self.announce_timing:
                _log.info("Request {} for {} lasted {} seconds".format(
                    request.id, request.method, (datetime.now() - start_time).total_seconds()))

            _log.debug("Sending client %s reply: %s", identity, reply)
            await self._socket.send_multipart([identity, *empty_frame, to_msgpack(reply)])
        except Exception as e:
            if self.serialize_exceptions:
                _log.exception('Exception thrown in _process_request')
            else:
                raise e




#!/usr/bin/env python

"""
WARNING: This file is auto-generated, do not edit by hand. See README.md.
"""

import sys

from warnings import warn
from rpcq._base import Message
from typing import Any, List, Dict, Optional

if sys.version_info < (3, 7):
    from rpcq.external.dataclasses import dataclass, field, InitVar
else:
    from dataclasses import dataclass, field, InitVar


@dataclass(eq=False, repr=False)
class ParameterSpec(Message):
    """
    Specification of a dynamic parameter type and array-length.
    """

    type: str = None
    """The parameter type, e.g., one of 'INTEGER', or 'FLOAT'."""

    length: int = 1
    """If this is not 1, the parameter is an array of this length."""


@dataclass(eq=False, repr=False)
class ParameterAref(Message):
    """
    A parametric expression.
    """

    name: str
    """The parameter name"""

    index: int
    """The array index."""


@dataclass(eq=False, repr=False)
class PatchTarget(Message):
    """
    Patchable memory location descriptor.
    """

    patch_type: ParameterSpec
    """Data type at this address."""

    patch_offset: int
    """Memory address of the patch."""


@dataclass(eq=False, repr=False)
class RPCRequest(Message):
    """
    A single request object according to the JSONRPC standard.
    """

    method: str
    """The RPC function name."""

    params: Any
    """The RPC function arguments."""

    id: str
    """RPC request id (used to verify that request and response belong together)."""

    jsonrpc: str = "2.0"
    """The JSONRPC version."""


@dataclass(eq=False, repr=False)
class RPCWarning(Message):
    """
    An individual warning emitted in the course of RPC processing.
    """

    body: str
    """The warning string."""

    kind: Optional[str] = None
    """The type of the warning raised."""


@dataclass(eq=False, repr=False)
class RPCReply(Message):
    """
    The reply for a JSONRPC request.
    """

    id: str
    """The RPC request id."""

    jsonrpc: str = "2.0"
    """The JSONRPC version."""

    result: Optional[Any] = None
    """The RPC result."""

    warnings: List[RPCWarning] = field(default_factory=list)
    """A list of warnings that occurred during request processing."""


@dataclass(eq=False, repr=False)
class RPCError(Message):
    """
    A error message for JSONRPC requests.
    """

    error: str
    """The error message."""

    id: str
    """The RPC request id."""

    jsonrpc: str = "2.0"
    """The JSONRPC version."""

    warnings: List[RPCWarning] = field(default_factory=list)
    """A list of warnings that occurred during request processing."""


@dataclass(eq=False, repr=False)
class TargetDevice(Message):
    """
    ISA and specs for a particular device.
    """

    isa: Dict[str, Dict]
    """Instruction-set architecture for this device."""

    specs: Dict[str, Dict]
    """Fidelities and coherence times for this device."""


@dataclass(eq=False, repr=False)
class RandomizedBenchmarkingRequest(Message):
    """
    RPC request payload for generating a randomized benchmarking sequence.
    """

    depth: int
    """Depth of the benchmarking sequence."""

    qubits: int
    """Number of qubits involved in the benchmarking sequence."""

    gateset: List[str]
    """List of Quil programs, each describing a Clifford."""

    seed: Optional[int] = None
    """PRNG seed. Set this to guarantee repeatable results."""

    interleaver: Optional[str] = None
    """Fixed Clifford, specified as a Quil string, to interleave through an RB sequence."""


@dataclass(eq=False, repr=False)
class RandomizedBenchmarkingResponse(Message):
    """
    RPC reply payload for a randomly generated benchmarking sequence.
    """

    sequence: List[List[int]]
    """List of Cliffords, each expressed as a list of generator indices."""


@dataclass(eq=False, repr=False)
class PauliTerm(Message):
    """
    Specification of a single Pauli term as a tensor product of Pauli factors.
    """

    indices: List[int]
    """Qubit indices onto which the factors of a Pauli term are applied."""

    symbols: List[str]
    """Ordered factors of a Pauli term."""


@dataclass(eq=False, repr=False)
class ConjugateByCliffordRequest(Message):
    """
    RPC request payload for conjugating a Pauli element by a Clifford element.
    """

    pauli: PauliTerm
    """Specification of a Pauli element."""

    clifford: str
    """Specification of a Clifford element."""


@dataclass(eq=False, repr=False)
class ConjugateByCliffordResponse(Message):
    """
    RPC reply payload for a Pauli element as conjugated by a Clifford element.
    """

    phase: int
    """Encoded global phase factor on the emitted Pauli."""

    pauli: str
    """Description of the encoded Pauli."""


@dataclass(eq=False, repr=False)
class NativeQuilRequest(Message):
    """
    Quil and the device metadata necessary for quilc.
    """

    quil: str
    """Arbitrary Quil to be sent to quilc."""

    target_device: TargetDevice
    """Specifications for the device to target with quilc."""


@dataclass(eq=False, repr=False)
class NativeQuilMetadata(Message):
    """
    Metadata for a native quil program.
    """

    final_rewiring: List[int] = field(default_factory=list)
    """Output qubit index relabeling due to SWAP insertion."""

    gate_depth: Optional[int] = None
    """Maximum number of successive gates in the native quil program."""

    gate_volume: Optional[int] = None
    """Total number of gates in the native quil program."""

    multiqubit_gate_depth: Optional[int] = None
    """Maximum number of successive two-qubit gates in the native quil program."""

    program_duration: Optional[float] = None
    """Rough estimate of native quil program length in nanoseconds."""

    program_fidelity: Optional[float] = None
    """Rough estimate of the fidelity of the full native quil program, uses specs."""

    topological_swaps: Optional[int] = None
    """Total number of SWAPs in the native quil program."""


@dataclass(eq=False, repr=False)
class NativeQuilResponse(Message):
    """
    Native Quil and associated metadata returned from quilc.
    """

    quil: str
    """Native Quil returned from quilc."""

    metadata: Optional[NativeQuilMetadata] = None
    """Metadata for the returned Native Quil."""


@dataclass(eq=False, repr=False)
class RewriteArithmeticRequest(Message):
    """
    A request type to handle compiling arithmetic out of gate parameters.
    """

    quil: str
    """Native Quil for which to rewrite arithmetic parameters."""


@dataclass(eq=False, repr=False)
class RewriteArithmeticResponse(Message):
    """
    The data needed to run programs with gate arithmetic on the hardware.
    """

    quil: str
    """Native Quil rewritten with no arithmetic in gate parameters."""

    original_memory_descriptors: Dict[str, ParameterSpec] = field(default_factory=dict)
    """The declared memory descriptors in the Quil of the related request."""

    recalculation_table: Dict[ParameterAref, str] = field(default_factory=dict)
    """A mapping from memory references to the original gate arithmetic."""


@dataclass(eq=False, repr=False)
class BinaryExecutableRequest(Message):
    """
    Native Quil and the information needed to create binary executables.
    """

    quil: str
    """Native Quil to be translated into an executable program."""

    num_shots: int
    """The number of times to repeat the program."""


@dataclass(eq=False, repr=False)
class BinaryExecutableResponse(Message):
    """
    Program to run on the QPU.
    """

    program: str
    """Execution settings and sequencer binaries."""

    memory_descriptors: Dict[str, ParameterSpec] = field(default_factory=dict)
    """Internal field for constructing patch tables."""

    ro_sources: List[Any] = field(default_factory=list)
    """Internal field for reshaping returned buffers."""


@dataclass(eq=False, repr=False)
class PyQuilExecutableResponse(Message):
    """
    rpcQ-serializable form of a pyQuil Program object.
    """

    program: str
    """String representation of a Quil program."""

    attributes: Dict[str, Any]
    """Miscellaneous attributes to be unpacked onto the pyQuil Program object."""


@dataclass(eq=False, repr=False)
class QPURequest(Message):
    """
    Program and patch values to send to the QPU for execution.
    """

    program: Any
    """Execution settings and sequencer binaries."""

    patch_values: Dict[str, List[Any]]
    """Dictionary mapping data names to data values for patching the binary."""

    id: str
    """QPU request ID."""





##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Class with json_rpc_call decorator for asynchronous JSON RPC calls
"""
import asyncio
import logging
import traceback
from typing import Union

from rpcq._utils import rpc_reply, rpc_error, RPCMethodError, get_input, \
                        catch_warnings
from rpcq.messages import RPCRequest, RPCReply, RPCError

_log = logging.getLogger(__name__)


class RPCSpec(object):
    """
    Class for keeping track of class methods that are exposed to the JSON RPC interface
    """
    def __init__(self, *, provide_tracebacks: bool = True, serialize_exceptions: bool = True):
        """
        Create a JsonRpcSpec object.

        Usage:
            jr = JsonRpcSpec()

            class MyClass(object):
                def __init__(self):
                    self.num = 5

            @jr.add_method
            def add(obj, *args):
                return sum(args) + obj.num

            obj = MyClass()

            request = {
                "jsonrpc": "2.0",
                "id": "0",
                "method": "add",
                "params": [1, 2]
            }

            reply = jr.call(request, obj)

        :param provide_tracebacks: If set to True, unhandled exceptions which occur during RPC call
            implementations will have their tracebacks forwarded to the calling client as part of
            the generated RPCError reply objject. If set to False, the generated RPCError reply will
            omit this information (but the traceback will still get written to the logfile).
        :param serialize_exceptions: If set to True, unhandled exceptions which occur during RPC
            call implementations will be serialized into RPCError messages (which the Server
            instance will then probably send to the corresponding Client).  If set to False, the
            exception is re-raised and left for the local caller to handle further.
        """
        self._json_rpc_methods = {}
        self.provide_tracebacks = provide_tracebacks
        self.serialize_exceptions = serialize_exceptions

    def add_handler(self, f):
        """
        Adds the function f to a dictionary of JSON RPC methods.

        :param callable f: Method to be exposed
        :return:
        """
        if f.__name__.startswith('rpc_'):
            raise ValueError("Server method names cannot start with rpc_.")
        self._json_rpc_methods[f.__name__] = f
        return f

    def get_handler(self, request):
        """
        Get callable from JSON RPC request

        :param RPCRequest request: JSON RPC request
        :return: Method
        :rtype: callable
        """
        try:
            f = self._json_rpc_methods[request.method]

        except (AttributeError, KeyError):  # pragma no coverage
            raise RPCMethodError("Received invalid method '{}'".format(request.method))

        return f

    async def run_handler(self, request: RPCRequest) -> Union[RPCReply, RPCError]:
        """
        Process a JSON RPC request

        :param RPCRequest request: JSON RPC request
        :return: JSON RPC reply
        """
        with catch_warnings(record=True) as warnings:
            try:
                rpc_handler = self.get_handler(request)
            except RPCMethodError as e:
                return rpc_error(request.id, str(e), warnings=warnings)

            try:
                # Run RPC and get result
                args, kwargs = get_input(request.params)
                result = rpc_handler(*args, **kwargs)

                if asyncio.iscoroutine(result):
                    result = await result

            except Exception as e:
                if self.serialize_exceptions:
                    _traceback = traceback.format_exc()
                    _log.error(_traceback)
                    if self.provide_tracebacks:
                        return rpc_error(request.id, "{}\n{}".format(str(e), _traceback),
                                         warnings=warnings)
                    else:
                        return rpc_error(request.id, str(e), warnings=warnings)
                else:
                    raise e

            return rpc_reply(request.id, result, warnings=warnings)



#!/usr/bin/env python

"""
WARNING: This file is auto-generated, do not edit by hand. See README.md.
"""

import sys

from warnings import warn
from rpcq._base import Message
from rpcq.messages import ParameterSpec, PatchTarget
from typing import Any, List, Dict, Optional

if sys.version_info < (3, 7):
    from rpcq.external.dataclasses import dataclass, field, InitVar
else:
    from dataclasses import dataclass, field, InitVar


@dataclass(eq=False, repr=False)
class Frame(Message):
    """
    A frame encapsulates any rotating frame
      relative to which control or readout waveforms may be defined.
    """

    direction: str
    """'rx' or 'tx'"""

    sample_rate: float
    """The sample rate [Hz] of the associated AWG/ADC"""

    frequency: float
    """The frame frequency [Hz]"""


@dataclass(eq=False, repr=False)
class Resources(Message):
    """
    The resources required by a job
    """

    qubits: List[str] = field(default_factory=list)
    """A list of qubits blocked/required by a job."""

    frames: Dict[str, Frame] = field(default_factory=dict)
    """RF/UHF frames by label."""

    frames_to_controls: Dict[str, str] = field(default_factory=dict)
    """Mapping of frames to control channels by labels."""


@dataclass(eq=False, repr=False)
class AbstractWaveform(Message):
    """
    A waveform envelope defined for a specific frame. This abstract class is made concrete by either a `Waveform` or a templated waveform such as `GaussianWaveform`
    """

    frame: str
    """The label of the associated tx-frame."""


@dataclass(eq=False, repr=False)
class Waveform(AbstractWaveform):
    """
    A waveform envelope defined by specific IQ values for a specific frame.
    """

    iqs: List[float] = field(default_factory=list)
    """The raw waveform envelope samples, alternating I and Q values."""


@dataclass(eq=False, repr=False)
class TemplateWaveform(AbstractWaveform):
    """
    A waveform envelope defined for a specific frame. A templated waveform is defined by a parameterized pulseshape rather than explicit IQ values. The message specification does not enforce that the duration is implementable on the hardware.
    """

    duration: float
    """Length of the pulse in seconds"""

    scale: float = 1.e+0
    """Scale to apply to waveform envelope"""

    phase: float = 0.0e+0
    """Phase [units of tau=2pi] to rotate the complex waveform envelope."""

    detuning: float = 0.0e+0
    """Modulation to apply to the waveform in Hz"""


@dataclass(eq=False, repr=False)
class GaussianWaveform(TemplateWaveform):
    """
    A Gaussian shaped waveform envelope defined for a specific frame.
    """

    fwhm: Optional[float] = None
    """Full Width Half Max shape paramter in seconds"""

    t0: Optional[float] = None
    """Center time coordinate of the shape in seconds. Defaults to mid-point of pulse."""


@dataclass(eq=False, repr=False)
class DragGaussianWaveform(TemplateWaveform):
    """
    A DRAG Gaussian shaped waveform envelope defined for a specific frame.
    """

    fwhm: Optional[float] = None
    """Full Width Half Max shape paramter in seconds"""

    t0: Optional[float] = None
    """Center time coordinate of the shape in seconds. Defaults to mid-point of pulse."""

    anh: float = -2.1e+8
    """Anharmonicity of the qubit, f01-f12 in (Hz)"""

    alpha: float = 0.0e+0
    """Dimensionless DRAG parameter"""


@dataclass(eq=False, repr=False)
class HermiteGaussianWaveform(TemplateWaveform):
    """
    Hermite-Gaussian shaped pulse. Reference: Effects of arbitrary laser
      or NMR pulse shapes on population inversion and coherence Warren S. Warren.
      81, (1984); doi: 10.1063/1.447644
    """

    fwhm: Optional[float] = None
    """Full Width Half Max shape paramter in seconds"""

    t0: float = None
    """Center time coordinate of the shape in seconds. Defaults to mid-point of pulse."""

    anh: float = -2.1e+8
    """Anharmonicity of the qubit, f01-f12 in Hz"""

    alpha: float = 0.0e+0
    """Dimensionless DRAG parameter"""

    second_order_hrm_coeff: float = 9.56e-1
    """Second order coefficient (see paper)"""


@dataclass(eq=False, repr=False)
class ErfSquareWaveform(TemplateWaveform):
    """
    Pulse with a flat top and rounded shoulders given by error functions
    """

    risetime: float = 1.e-9
    """The width of the rise and fall sections in seconds."""

    pad_left: float = 0.0e+0
    """Length of zero-amplitude padding before the pulse in seconds."""

    pad_right: float = 0.0e+0
    """Length of zero-amplitude padding after the pulse in seconds."""


@dataclass(eq=False, repr=False)
class FlatWaveform(TemplateWaveform):
    """
    Flat pulse.
    """

    iq: List[float] = field(default_factory=list)
    """Individual IQ point to hold constant"""


@dataclass(eq=False, repr=False)
class AbstractKernel(Message):
    """
    An integration kernel defined for a specific frame. This abstract class is made concrete by either a `FilterKernel` or `TemplateKernel`
    """

    frame: str
    """The label of the associated rx-frame."""


@dataclass(eq=False, repr=False)
class FilterKernel(AbstractKernel):
    """
    A filter kernel to produce scalar readout features from acquired readout waveforms.
    """

    iqs: List[float] = field(default_factory=list)
    """The raw kernel coefficients, alternating real and imaginary parts."""

    bias: float = 0.0e+0
    """The kernel is offset by this real value. Can be used to ensure the decision threshold lies at 0.0."""


@dataclass(eq=False, repr=False)
class TemplateKernel(AbstractKernel):
    """
    An integration kernel defined for a specific frame.
    """

    duration: float
    """Length of the boxcar kernel in seconds"""

    bias: float = 0.0e+0
    """The kernel is offset by this real value. Can be used to ensure the decision threshold lies at 0.0."""

    scale: float = 1.e+0
    """Scale to apply to boxcar kernel"""

    phase: float = 0.0e+0
    """Phase [units of tau=2pi] to rotate the kernel by."""

    detuning: float = 0.0e+0
    """Modulation to apply to the filter kernel in Hz"""


@dataclass(eq=False, repr=False)
class FlatKernel(TemplateKernel):
    """
    An unnormalized flat or boxcar integration kernel.
    """


@dataclass(eq=False, repr=False)
class BoxcarAveragerKernel(TemplateKernel):
    """
    A normalized flat or boxcar integration kernel.
    """


@dataclass(eq=False, repr=False)
class FilterNode(Message):
    """
    A node in the filter pipeline.
    """

    module: str
    """Absolute python module import path in which the filter
          class is defined."""

    filter_type: str
    """The type (class name) of the filter."""

    source: str
    """Filter node label of the input to this node."""

    publish: bool
    """If True, return the output of this node with the job
          results (and publish a stream for it)."""

    params: Dict[str, float] = field(default_factory=dict)
    """Additional filter parameters."""


@dataclass(eq=False, repr=False)
class DataAxis(Message):
    """
    A data axis allows to label element(s) of a stream.
    """

    name: str
    """Label for the axis, e.g., 'time' or 'shot_index'."""

    points: List[float] = field(default_factory=list)
    """The sequence of values along the axis."""


@dataclass(eq=False, repr=False)
class Receiver(Message):
    """
    The receiver settings generated by the low-level
      translator.
    """

    instrument: str
    """The instrument name"""

    channel: str
    """The instrument channel (label)"""

    stream: str
    """Name of the associated (raw) output stream that
          should be published."""

    publish: bool
    """Whether to publish the raw output stream."""

    data_axes: List[DataAxis] = field(default_factory=list)
    """Ordered list of DataAxis objects that together
          uniquely label each element in the stream."""


@dataclass(eq=False, repr=False)
class Program(Message):
    """
    The dynamic aspects (waveforms, readout kernels, scheduled
  instructions and parameters) of a job.
    """

    waveforms: Dict[str, AbstractWaveform] = field(default_factory=dict)
    """The waveforms appearing in the program by waveform
          label."""

    filters: Dict[str, AbstractKernel] = field(default_factory=dict)
    """The readout filter kernels appearing in the program by
          feature label."""

    scheduled_instructions: List[Dict] = field(default_factory=list)
    """The ordered sequence scheduled instruction objects."""

    parameters: Dict[str, ParameterSpec] = field(default_factory=dict)
    """A mapping of dynamic parameter names to their type
          specification."""


@dataclass(eq=False, repr=False)
class ScheduleIRJob(Message):
    """
    The unit of work to be executed.
    """

    num_shots: int
    """How many repetitions the job should be executed for."""

    resources: Resources
    """The resources required by the job."""

    program: Program
    """The actual program to be executed."""

    operating_point: Dict[str, Dict] = field(default_factory=dict)
    """Operating points or static instrument channel settings
          (mapping control_name (instrument name) -> instrument channel settings
          (instrument settings) dictionary)."""

    filter_pipeline: Dict[str, FilterNode] = field(default_factory=dict)
    """The filter pipeline. Mapping of node labels to
          FilterNode's."""

    job_id: InitVar[Optional[str]] = None
    """A unique ID to help the submitter track the job."""

    def _extend_by_deprecated_fields(self, d):
        super()._extend_by_deprecated_fields(d)

    def __post_init__(self, job_id):
        if job_id is not None:
            warn('job_id is deprecated; please don\'t set it anymore')


@dataclass(eq=False, repr=False)
class ParameterExpression(Message):
    """
    A parametric expression.
    """

    operator: str
    """The operator '+', '-', '*'. The operands can be
          constant floating point numbers or strings referencing a dynamic
          program parameter or a ParameterAref to index into an array or
          itself a ParameterExpression."""

    a: Any
    """The first operand"""

    b: Any
    """The second operand"""


@dataclass(eq=False, repr=False)
class Instruction(Message):
    """
    An instruction superclass.
    """

    time: float
    """The time at which the instruction is emitted [in seconds]."""


@dataclass(eq=False, repr=False)
class DebugMessage(Instruction):
    """
    Instructs the target to emit a specified debug message.
    """

    frame: str
    """The frame label that owns this debug message."""

    message: int
    """The 2-byte wide debug message to emit."""


@dataclass(eq=False, repr=False)
class Pulse(Instruction):
    """
    Instruction to play a pulse with some (modified) waveform
      envelope at a specific time on a specific frame.
    """

    frame: str
    """The tx-frame label on which the pulse is played."""

    waveform: str
    """The waveform label"""

    scale: float = 1.e+0
    """Dimensionless (re-)scaling factor which is applied to
          the envelope."""

    phase: float = 0.0e+0
    """Static phase angle [units of tau=2pi] by which the
          envelope quadratures are rotated."""

    detuning: float = 0.0e+0
    """Detuning [Hz] with which the pulse envelope should be
          modulated relative to the frame frequency."""


@dataclass(eq=False, repr=False)
class FlatPulse(Instruction):
    """
    Instruction to play a pulse with a constant amplitude
      (except for phase modulation) at a specific time on a specific frame.
    """

    frame: str
    """The tx-frame label on which the pulse is played."""

    iq: List[float]
    """The I and Q value of the constant pulse."""

    duration: float
    """The duration of the pulse in [seconds], should be a
          multiple of the associated tx-frame's inverse sample rate."""

    phase: float = 0.0e+0
    """Static phase angle [units of tau=2pi] by which the
          envelope quadratures are rotated."""

    detuning: float = 0.0e+0
    """Detuning [Hz] with which the pulse envelope should be
          modulated relative to the frame frequency."""

    scale: float = 1.e+0
    """Dimensionless (re-)scaling factor which is applied to
          the envelope."""


@dataclass(eq=False, repr=False)
class SetPhase(Instruction):
    """
    Set the phase of a frame to an absolute value at a specific
      time.
    """

    frame: str
    """The frame label for which to set the phase."""

    phase: float = 0.0e+0
    """Phase angle [units of tau=2pi] to update the frame phase
          to."""


@dataclass(eq=False, repr=False)
class ShiftPhase(Instruction):
    """
    Shift the phase of a frame by a relative value at a
      specific time.
    """

    frame: str
    """The frame label for which to set the phase."""

    delta: Any = None
    """Phase angle [units of tau=2pi] by which to shift the
          frame phase.  Can be a numerical value, a ParameterExpression or a
          ParameterAref."""


@dataclass(eq=False, repr=False)
class SwapPhases(Instruction):
    """
    Swap the phases of two tx-frames at a specific time.
    """

    frame_a: str
    """The first frame's label."""

    frame_b: str
    """The second frame's label."""


@dataclass(eq=False, repr=False)
class SetFrequency(Instruction):
    """
    Set the frequency of a tx-frame to a specific value at a
      specific time.
    """

    frame: str
    """The frame label for which to set the frequency."""

    frequency: float = 0.0e+0
    """The frequency [Hz] to set the frame frequency to."""


@dataclass(eq=False, repr=False)
class ShiftFrequency(Instruction):
    """
    Shift the frequency of a tx-frame by a specific amount at a
      specific time.
    """

    frame: str
    """The frame label for which to set the frequency."""

    delta: float = 0.0e+0
    """Frequency shift (new-old) [Hz] to apply to the frame
          frequency."""


@dataclass(eq=False, repr=False)
class Capture(Instruction):
    """
    Specify an acquisition on an rx-frame as well as the
      filters to apply.
    """

    frame: str
    """The rx-frame label on which to trigger the acquisition."""

    duration: float
    """The duration of the acquisition in [seconds]"""

    filters: List[str] = field(default_factory=list)
    """An ordered list of labels of filter kernels to apply to
          the captured waveform."""

    send_to_host: bool = True
    """Transmit the readout bit back to Lodgepole.
          (Unnecessary for fully calibrated active reset captures)."""

    phase: float = 0.0e+0
    """Static phase angle [units of tau=2pi] by which the
          envelope quadratures are rotated."""

    detuning: float = 0.0e+0
    """Detuning [Hz] with which the pulse envelope should be
          modulated relative to the frame frequency."""


@dataclass(eq=False, repr=False)
class RackMeta(Message):
    """
    Meta information about a rack configuration.
    """

    rack_id: Optional[str] = None
    """A unique identifier for the rack."""

    rack_version: Optional[int] = None
    """A version of the rack configuration."""

    schema_version: int = 0
    """A version of the rack configuration."""


@dataclass(eq=False, repr=False)
class QPU(Message):
    """
    Configuration info for the QPU
    """

    chip_label: str
    """The fabrication label for the QPU chip."""

    qubits: List[str] = field(default_factory=list)
    """A list of qubits labels."""

    controls: Dict[str, List] = field(default_factory=dict)
    """A mapping of control labels to tuples (instrument
          label, channel label)."""

    controls_by_qubit: Dict[str, List] = field(default_factory=dict)
    """A map of qubit label to list of controls that should be
          considered blocked when the qubit is part of a job execution."""


@dataclass(eq=False, repr=False)
class Instrument(Message):
    """
    Instrument settings.
    """

    address: str
    """The full address of a QPU."""

    module: str
    """Full python import path for the module that includes
          the instrument driver."""

    instrument_type: str
    """Instrument type (driver class name)."""

    channels: Dict[str, Any] = field(default_factory=dict)
    """Mapping of channel labels to channel settings"""

    virtual: bool = False
    """Whether the instrument is virtual."""

    setup: Dict[str, Any] = field(default_factory=dict)
    """Any additional information used by the instrument for
          one-time-setup"""


@dataclass(eq=False, repr=False)
class DeployedRack(Message):
    """
    The rack configuration for lodgepole.
    """

    rack_meta: RackMeta
    """Meta information about the deployed rack."""

    qpu: QPU
    """Information about the QPU."""

    instruments: Dict[str, Instrument] = field(default_factory=dict)
    """Mapping of instrument name to instrument settings."""


@dataclass(eq=False, repr=False)
class AWGChannel(Message):
    """
    Configuration of a single RF channel.
    """

    sample_rate: float
    """The sampling rate [Hz] of the associated DAC/ADC
          component."""

    direction: str
    """'rx' or 'tx'"""

    lo_frequency: Optional[float] = None
    """The local oscillator frequency [Hz] of the channel."""

    gain: Optional[float] = None
    """If there is an amplifier, the amplifier gain [dB]."""

    delay: float = 0.0e+0
    """Delay [seconds] to account for inter-channel skew."""


@dataclass(eq=False, repr=False)
class QFDChannel(Message):
    """
    Configuration for a single QFD Channel.
    """

    channel_index: int
    """The channel index on the QFD, zero indexed from the
          lowest channel, as installed in the box."""

    direction: Optional[str] = "tx"
    """The QFD is a device that transmits pulses."""

    nco_frequency: Optional[float] = 0.0e+0
    """The DAC NCO frequency [Hz]."""

    gain: Optional[float] = 0.0e+0
    """The output gain on the DAC in [dB]. Note that this
          should be in the range -45dB to 0dB and is rounded to the
          nearest 3dB step."""

    delay: float = 0.0e+0
    """Delay [seconds] to account for inter-channel skew."""

    flux_current: Optional[float] = None
    """Slow flux current [Amps]."""


@dataclass(eq=False, repr=False)
class QGSChannel(Message):
    """
    Configuration for a single QGS Channel.
    """

    direction: Optional[str] = "tx"
    """The QGS is a device that transmits pulses."""

    nco_frequency: Optional[float] = 2.e+9
    """The DAC NCO frequency [Hz]."""

    gain: Optional[float] = 0.0e+0
    """The output gain on the DAC in [dB]. Note that this
          should be in the range -45dB to 0dB and is rounded to the
          nearest 3dB step."""

    delay: float = 0.0e+0
    """Delay [seconds] to account for inter-channel skew."""


@dataclass(eq=False, repr=False)
class QRTChannel(Message):
    """
    Configuration for a single QRT Channel.
    """

    direction: Optional[str] = "tx"
    """The QRT is a device that transmits readout pulses."""

    nco_frequency: Optional[float] = 1.25e+9
    """The DAC NCO frequency [Hz]."""

    gain: Optional[float] = 0.0e+0
    """The output gain on the DAC in [dB]. Note that this should be in the range
       -45dB to 0dB and is rounded to the nearest 3dB step."""

    delay: float = 0.0e+0
    """Delay [seconds] to account for inter-channel skew."""


@dataclass(eq=False, repr=False)
class QRRChannel(Message):
    """
    Configuration for a single QRR Channel.
    """

    channel_index: int
    """The channel index on the QRR, zero indexed from the lowest channel,
        as installed in the box."""

    direction: Optional[str] = "rx"
    """The QRR is a device that receives readout pulses."""

    nco_frequency: Optional[float] = 0.0e+0
    """The ADC NCO frequency [Hz]."""

    gain: Optional[float] = 0.0e+0
    """The input gain on the ADC in [dB]. Note that this should be in the range
       -45dB to 0dB and is rounded to the nearest 3dB step."""

    delay: float = 0.0e+0
    """Delay [seconds] to account for inter-channel skew."""


@dataclass(eq=False, repr=False)
class LegacyUSRPSequencer(Message):
    """
    Configuration for a Legacy USRP Sequencer
    """

    tx_channel: Optional[str] = None
    """The label of the associated tx channel."""

    rx_channel: Optional[str] = None
    """The label of the associated rx channel."""


@dataclass(eq=False, repr=False)
class QFDSequencer(Message):
    """
    Configuration for a single QFD Sequencer.
    """

    tx_channel: str
    """The label of the associated channel."""


@dataclass(eq=False, repr=False)
class QGSSequencer(Message):
    """
    Configuration for a single QGS Sequencer.
    """

    tx_channel: str
    """The label of the associated channel."""


@dataclass(eq=False, repr=False)
class QRTSequencer(Message):
    """
    Configuration for a single readout transmit (QRT) sequencer.
    """

    tx_channel: str
    """The label of the associated tx channel."""

    sequencer_index: int
    """The sequencer index (0-7) of this sequencer."""

    low_freq_range: Optional[bool] = False
    """Used to signal if this sequencer is in the low frequency configuration."""


@dataclass(eq=False, repr=False)
class QRTx2Sequencer(Message):
    """
    Configuration for a dual readout transmit (QRTx2) sequencer.
    """

    tx_channel: str
    """The label of the associated tx channel."""

    sequencer_index: int
    """The sequencer index (0-15) of this sequencer."""

    low_freq_range: Optional[bool] = False
    """Used to signal if this sequencer is in the low frequency configuration."""


@dataclass(eq=False, repr=False)
class QRRSequencer(Message):
    """
    Configuration for a single readout receive (QRR) sequencer.
    """

    rx_channel: str
    """The label of the associated rx channel."""

    sequencer_index: int
    """The sequencer index (0-15) to assign. Note that only
         sequencer 0 can return raw readout measurements."""


@dataclass(eq=False, repr=False)
class USICardSequencer(Message):
    """
    Configuration for the card which
      interfaces with the USI Target on the USRP.
    """

    tx_channel: str
    """The label of the associated channel."""


@dataclass(eq=False, repr=False)
class USITargetSequencer(Message):
    """
    Configuration for a single USITarget Sequencer.
    """

    tx_channel: str
    """The label of the associated intial tx channel."""

    rx_channel: str
    """The label of the associated initial rx channel."""

    sequencer_index: int
    """The sequencer index (0-7) to assign. Note that only
           sequencer 0 has the ability to use the NCO or capture raw readout
           streams."""


@dataclass(eq=False, repr=False)
class CWFrequencySweep(Message):
    """
    Configuration of a continuous wave frequency sweep.
    """

    start: float
    """Start frequency of the sweep, in Hz"""

    stop: float
    """Stop frequency of the sweep, in Hz"""

    num_pts: int
    """Number of frequency points to sample, cast to int."""

    source: int
    """Source port number"""

    measure: int
    """Measure port number"""


@dataclass(eq=False, repr=False)
class VNASettings(Message):
    """
    Configuration of VNA settings for a continuous wave sweep.
    """

    e_delay: float
    """Electrical delay in seconds from source to measure port"""

    phase_offset: float
    """Phase offset in degrees from measured to reported phase"""

    bandwidth: float
    """Bandwidth of the sweep, in Hz"""

    power: float
    """Source power in dBm"""

    freq_sweep: CWFrequencySweep
    """Frequency sweep settings"""

    averaging: int = 1
    """Sets the number of points to combine into an averaged
          trace"""


@dataclass(eq=False, repr=False)
class TimeBomb(Message):
    """
    Payload used to match a job with a particular execution
      target.
    """

    deadline: str
    """Deadline, specified in the format
          '%Y-%m-%dT%H:%M:%S.000Z', after which this job becomes unexecutable."""

    chip_label: str
    """Label string for the chip on which this job is meant to
          execute."""


@dataclass(eq=False, repr=False)
class MicrowaveSourceSettings(Message):
    """
    Configuration of Microwave Source settings for operating amplifiers.
    """

    frequency: float
    """Frequency setting for microwave source (Hz)."""

    power: float
    """Power setting for microwave source (dBm)."""

    output: bool
    """Output setting for microwave source. If true, the source will be turned on."""


@dataclass(eq=False, repr=False)
class ExecutorJob(Message):
    """
    Job which is sent directly to the executor
    """

    instrument_settings: Dict[str, Any]
    """Dict mapping instrument names to arbitrary instrument
          settings."""

    filter_pipeline: Dict[str, FilterNode]
    """The filter pipeline to process measured data."""

    receivers: Dict[str, Receiver]
    """Dict mapping stream names to receiver settings."""

    duration: Optional[float] = None
    """The total duration of the program execution in seconds."""

    timebomb: Optional[TimeBomb] = None
    """An optional payload used to match this job with a
          particular execution target."""


@dataclass(eq=False, repr=False)
class PatchableBinary(Message):
    """
    Tsunami binary with patching metadata for classical
      parameter modification.
    """

    base_binary: Any
    """Raw Tsunami binary object."""

    patch_table: Dict[str, PatchTarget]
    """Dictionary mapping patch names to their memory
          descriptors."""


@dataclass(eq=False, repr=False)
class ActiveReset(Message):
    """
    An active reset control sequence consisting of a repeated
      sequence of a measurement block and a feedback block conditional on the
      outcome of a specific measurement bit.  Regardless of the measurement
      outcomes the total duration of the control sequence is [attempts x
      (measurement_duration + feedback_duration)].  The total
      measurement_duration must be chosen to allow for enough time after any
      Capture commands for the measurement bit to propagate back to the gate
      cards that are actuating the feedback.
    """

    time: float
    """Time at which the ActiveReset begins in [seconds]."""

    measurement_duration: float
    """The duration of measurement block in [seconds]. The
          measurement bit is expected to have arrived on the QGS after
          this time relative to the overall start of the ActiveReset block."""

    feedback_duration: float
    """The duration of feedback block in [seconds]"""

    measurement_bit: int
    """The address of the readout bit to condition the
          feedback on.  The bit is first accessed after measurement_duration
          has elapsed."""

    attempts: int = 3
    """The number of times to repeat the active reset sequence."""

    measurement_instructions: List[Dict] = field(default_factory=list)
    """The ordered sequence of scheduled measurement
          instructions."""

    apply_feedback_when: bool = True
    """Apply the feedback when the measurement_bit equals the
          value of this flag."""

    feedback_instructions: List[Dict] = field(default_factory=list)
    """The ordered sequence of scheduled feedback instructions."""



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import asyncio
import logging
import time
from typing import Dict, Union
from warnings import warn

import zmq
import zmq.asyncio

from rpcq._base import to_msgpack, from_msgpack
import rpcq._utils as utils
from rpcq.messages import RPCError, RPCReply

_log = logging.getLogger(__name__)


class Client:
    """
    Client that executes methods on a remote server by sending JSON RPC requests to a socket.
    """
    def __init__(self, endpoint: str, timeout: float = None):
        """
        Create a client that connects to a server at <endpoint>.

        :param str endpoint: Socket endpoint, e.g. "tcp://localhost:1234"
        :param float timeout: Timeout in seconds for Server response, set to None to disable the timeout
        """
        # TODO: leaving self.timeout for backwards compatibility; we should move towards using rpc_timeout only
        self.timeout = timeout
        self.rpc_timeout = timeout
        self.endpoint = endpoint

        self._socket = self._connect_to_socket(zmq.Context(), endpoint)
        # The async socket can't be created yet because it's possible that the current event loop during Client creation
        # is different to the one used later to call a method, so we need to create the socket after the first call and
        # then cache it
        self._async_socket_cache = None

        # Mapping from request id to an event used to wake up the call that's waiting on that request.
        # This is necessary to support parallel, asynchronous calls where we don't know which
        # receive task will receive which reply.
        self._events: Dict[str, asyncio.Event] = {}

        # Cache of replies so that different tasks can share results with each other
        self._replies: Dict[str, Union[RPCReply, RPCError]] = {}

    def __setattr__(self, key, value):
        """
        Ensure rpc_timeout attribute gets update with timeout. Currently keeping self.timeout and
        self.rpc_timeout for backwards compatibility. We should move towards using rpc_timeout only.

        :param key: attribute key
        :param value: attribute value
        :return:
        """
        if key == 'timeout':
            self.rpc_timeout = value
        super().__setattr__(key, value)

    async def call_async(self, method_name: str, *args, rpc_timeout: float = None, **kwargs):
        """
        Send JSON RPC request to a backend socket and receive reply (asynchronously)

        :param method_name: Method name
        :param args: Args that will be passed to the remote function
        :param float rpc_timeout: Timeout in seconds for Server response, set to None to disable the timeout
        :param kwargs: Keyword args that will be passed to the remote function
        """
        # if an rpc_timeout override is not specified, use the one set in the Client attributes
        if rpc_timeout is None:
            rpc_timeout = self.rpc_timeout

        if rpc_timeout:
            # Implementation note: this simply wraps the call in a timeout and converts to the built-in TimeoutError
            try:
                return await asyncio.wait_for(self._call_async(method_name, *args, **kwargs), timeout=rpc_timeout)
            except asyncio.TimeoutError:
                raise TimeoutError(f"Timeout on client {self.endpoint}, method name {method_name}, class info: {self}")
        else:
            return await self._call_async(method_name, *args, **kwargs)

    async def _call_async(self, method_name: str, *args, **kwargs):
        """
        Sends a request to the socket and then wait for the reply.

        To deal with multiple, asynchronous requests we do not expect that the receive reply task
        scheduled from this call is the one that receives this call's reply and instead rely on
        Events to signal across multiple _async_call/_recv_reply tasks.
        """
        request = utils.rpc_request(method_name, *args, **kwargs)
        _log.debug("Sending request: %s", request)

        # setup an event to notify us when the reply is received (potentially by a task scheduled by
        # another call to _async_call). we do this before we send the request to catch the case
        # where the reply comes back before we re-enter this thread
        self._events[request.id] = asyncio.Event()

        # schedule a task to receive the reply to ensure we have a task to receive the reply
        asyncio.ensure_future(self._recv_reply())

        await self._async_socket.send_multipart([to_msgpack(request)])
        await self._events[request.id].wait()

        reply = self._replies.pop(request.id)
        if isinstance(reply, RPCError):
            raise utils.RPCError(reply.error)
        else:
            return reply.result

    async def _recv_reply(self):
        """
        Helper task to recieve a reply store the result and trigger the associated event.
        """
        raw_reply, = await self._async_socket.recv_multipart()
        reply = from_msgpack(raw_reply)
        _log.debug("Received reply: %s", reply)
        self._replies[reply.id] = reply
        self._events.pop(reply.id).set()

    def call(self, method_name: str, *args, rpc_timeout: float = None, **kwargs):
        """
        Send JSON RPC request to a backend socket and receive reply
        Note that this uses the default event loop to run in a blocking manner. If you would rather run in an async
        fashion or provide your own event loop then use .async_call instead

        :param method_name: Method name
        :param args: Args that will be passed to the remote function
        :param float rpc_timeout: Timeout in seconds for Server response, set to None to disable the timeout
        :param kwargs: Keyword args that will be passed to the remote function
        """
        request = utils.rpc_request(method_name, *args, **kwargs)
        _log.debug("Sending request: %s", request)

        self._socket.send_multipart([to_msgpack(request)])

        # if an rpc_timeout override is not specified, use the one set in the Client attributes
        if rpc_timeout is None:
            rpc_timeout = self.rpc_timeout

        start_time = time.time()
        while True:
            # Need to keep track of timeout manually in case this loop runs more than once. We subtract off already
            # elapsed time from the timeout. The call to max is to make sure we don't send a negative value
            # which would throw an error.
            timeout = max((start_time + rpc_timeout - time.time()) * 1000, 0) if rpc_timeout is not None else None
            if self._socket.poll(timeout) == 0:
                raise TimeoutError(f"Timeout on client {self.endpoint}, method name {method_name}, class info: {self}")

            raw_reply, = self._socket.recv_multipart()
            reply = from_msgpack(raw_reply)
            _log.debug("Received reply: %s", reply)

            # there's a possibility that the socket will have some leftover replies from a previous
            # request on it if that .call() was cancelled or timed out. Therefore, we need to discard replies that
            # don't match the request just like in the call_async case.
            if reply.id == request.id:
                break
            else:
                _log.debug('Discarding reply: %s', reply)

        for warning in reply.warnings:
            warn(f"{warning.kind}: {warning.body}")

        if isinstance(reply, RPCError):
            raise utils.RPCError(reply.error)
        else:
            return reply.result

    def close(self):
        """
        Close the sockets
        """
        self._socket.close()
        if self._async_socket_cache:
            self._async_socket_cache.close()
            self._async_socket_cache = None

    def _connect_to_socket(self, context: zmq.Context, endpoint: str):
        """
        Connect to a DEALER socket at endpoint and turn off lingering.

        :param context: ZMQ Context to use (potentially async)
        :param endpoint: Endpoint
        :return: Connected socket
        """
        socket = context.socket(zmq.DEALER)
        socket.connect(endpoint)
        socket.setsockopt(zmq.LINGER, 0)
        _log.debug("Client connected to endpoint %s", self.endpoint)
        return socket

    @property
    def _async_socket(self):
        """
        Creates a new async socket if one doesn't already exist for this Client
        """
        if not self._async_socket_cache:
            self._async_socket_cache = self._connect_to_socket(zmq.asyncio.Context(), self.endpoint)

        return self._async_socket_cache



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import sys

from setuptools import setup

if sys.version_info < (3, 6):
    raise ImportError('The rpcq library requires Python 3.6 or above.')

with open('VERSION.txt', 'r') as f:
    __version__ = f.read().strip()

with open('README.md', 'r') as f:
    long_description = f.read()

# save the source code in version.py
with open('rpcq/version.py', 'r') as f:
    version_file_source = f.read()

# overwrite version.py in the source distribution
with open('rpcq/version.py', 'w') as f:
    f.write(f'__version__ = \'{__version__}\'\n')

setup(
    name='rpcq',
    version=__version__,
    author='Rigetti Computing',
    author_email='info@rigetti.com',
    license='Apache-2.0',
    packages=[
        'rpcq',
        'rpcq.external'
    ],
    url='https://github.com/rigetticomputing/rpcq.git',
    description='''The RPC framework and message specification for Rigetti QCS.''',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'future',
        'msgpack>=0.6',
        'python-rapidjson',
        'pyzmq>=17',
        'ruamel.yaml',
        'typing'
    ],
    keywords='quantum rpc qcs',
    python_requires='>=3.6',
)

# restore version.py to its previous state
with open('rpcq/version.py', 'w') as f:
    f.write(version_file_source)



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import inspect
import sys

import msgpack
import rapidjson
from ruamel import yaml

if sys.version_info < (3, 7):
    from rpcq.external.dataclasses import astuple, replace, fields
else:
    from dataclasses import astuple, replace, fields


MAX_BIN_LEN = MAX_STR_LEN = 2 ** 31 - 1  # less than 2 GB
REPR_LIST_TRUNCATION = 10
"Number of list elements to print when calling repr on a Message with a list field."


def repr_value(value):
    """
    Represent a value in human readable form. For long list's this truncates the printed
    representation.

    :param value: The value to represent.
    :return: A string representation.
    :rtype: basestring
    """
    if isinstance(value, list) and len(value) > REPR_LIST_TRUNCATION:
        return "[{},...]".format(", ".join(map(repr, value[:REPR_LIST_TRUNCATION])))
    else:
        return repr(value)


class UnknownMessageType(Exception):
    """Raised when trying to decode an unknown message type."""


class Message:
    """
    Base class for messages.
    """

    def asdict(self):
        """
        Create a dictionary ``{fieldname1: fieldvalue1, ...}`` of the Message object.

        :return: A dictionary representation of the message.
        :rtype: Dict[str,Any]
        """
        return self.__dict__.copy()

    def astuple(self):
        """
        Create a tuple ``{fieldvalue1, ...}`` of the Message object.
         :return: A tuple representation of the message.
        :rtype: Tuple[Any]
        """
        return tuple(getattr(self, f.name) for f in fields(self))

    def replace(self, **kwargs):
        """
        Return a copy of the message object where the fields given in kwargs are
        replaced.

        :param kwargs: The replaced fields.
        :return: A copy of self.
        """
        return replace(self, **kwargs)

    def _extend_by_deprecated_fields(self, d):
        pass

    copy = replace

    def items(self):
        return self.__dict__.items()

    def get(self, key, default):
        return self.__dict__.get(key, default)

    def __getitem__(self, item):
        return self.__dict__[item]

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join("{}={}".format(k, repr_value(v))
                      for k, v in sorted(self.asdict().items(), key=lambda kv: kv[0])))

    def __eq__(self, other):
        return type(self) == type(other) and astuple(self) == astuple(other)

    def __hash__(self):
        return hash((self.__class__, astuple(self)))

    _types = None

    @staticmethod
    def types():
        """
        Return a mapping ``{type_name: (message_type, args)}`` for all defined Message's,
        where ``args`` is a list of kwarg names that message_type's __init__ function accepts.

        :return: A dictionary of ``Message`` types.
        :rtype: Dict[str,type]
        """
        if Message._types is None:
            Message._types = {}
            classes_to_process = [Message]
            while classes_to_process:
                atom = classes_to_process.pop()
                classes_to_process += atom.__subclasses__()
                Message._types[atom.__name__] = (atom, inspect.getfullargspec(atom.__init__).args)
        return Message._types


def _default(obj):
    if isinstance(obj, Message):
        d = obj.__dict__
        obj._extend_by_deprecated_fields(d)
        d["_type"] = obj.__class__.__name__
        return d
    else:
        raise TypeError('Object of type {} is not JSON serializable'.format(obj.__class__.__name__))


def _object_hook(obj):
    if '_type' in obj:
        try:
            class_dict = Message.types()
            msg_type, allowed_args = class_dict[obj['_type']]
        except KeyError:  # pragma no coverage
            raise UnknownMessageType("The message type {} is unknown".format(obj["_type"]))

        itms = {k: v for k, v in obj.items() if k in allowed_args}
        return msg_type(**itms)
    else:
        return obj


def to_msgpack(obj):
    """
    Convert Python objects (including rpcq objects) to a msgpack byte array
    :rtype: bytes
    """
    # Docs for `use_bin_type` parameter are somewhat hard to find so they're copied here:
    #   Use bin type introduced in msgpack spec 2.0 for bytes.
    #   It also enables str8 type for unicode.
    return msgpack.dumps(obj, default=_default, use_bin_type=True)


def from_msgpack(b, *, max_bin_len=MAX_BIN_LEN, max_str_len=MAX_STR_LEN):
    """
    Convert a msgpack byte array into Python objects (including rpcq objects)
    """
    # Docs for raw parameter are somewhat hard to find so they're copied here:
    #   If true, unpack msgpack raw to Python bytes (default).
    #   Otherwise, unpack to Python str (or unicode on Python 2) by decoding with UTF-8 encoding (recommended).
    #   In msgpack >= 0.6, max_xxx_len is reduced from 2 GB to 1 MB, so we set the relevant ones
    #       to 2 GB as to not run into issues with the size of the values returned from rpcq
    return msgpack.loads(b, object_hook=_object_hook, raw=False,
                         max_bin_len=max_bin_len, max_str_len=max_str_len)


def to_json(obj):
    """
    Convert Python objects (including rpcq objects) to a JSON string.
    :rtype: str
    """
    return rapidjson.dumps(obj, default=_default)


def from_json(s):
    """
    Convert a JSON string into Python objects (including rpcq objects).
    """
    return rapidjson.loads(s, object_hook=_object_hook)


def to_yaml_file(obj, f):
    """
    Convert Python objects (including rpcq messages) to yaml and write it to `f`.
    """
    yaml.dump(rapidjson.loads(to_json(obj)), f)


def from_yaml_file(f):
    """
    Read a yaml file and convert to Python objects (including rpcq messages).
    """
    return from_json(to_json(yaml.load(f, Loader=yaml.Loader)))
